#include "globals.h"

void iconActivateCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void listItemActivateCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void readDir(
#if NeedFunctionPrototypes
        struct FileManager*, int, char*, char*, struct stat*, int, int, char*
#endif
);
void reloadDir(
#if NeedFunctionPrototypes
        struct FileManager*, int
#endif
);
void loadDirStruct(
#if NeedFunctionPrototypes
        struct FileManager*, int, int*, int
#endif
);
void deleteDirStruct(
#if NeedFunctionPrototypes
        int, struct FileManager*
#endif
);
void setScrollbar(
#if NeedFunctionPrototypes
        struct FileManager*, int
#endif
);
void sortFiles(
#if NeedFunctionPrototypes
        struct DirInfo*
#endif
);
void chooseCB(
#if NeedFunctionPrototypes
        Widget, XtPointer, XtPointer
#endif
);
void addListObj(
#if NeedFunctionPrototypes
        struct FileManager*
#endif
);
void resizeCB(
#if NeedFunctionPrototypes
        Widget, XEvent*
#endif
);
void cmdPopup(
#if NeedFunctionPrototypes
        Widget, XEvent*
#endif
);


static XtTranslations parsed_resize_xlations = (XtTranslations)NULL;
static XtTranslations parsed_cmd_xlations = (XtTranslations)NULL;
static char resizeTranslations[] = "#augment <Configure> : resizeReq()";
static char cmdTranslations[] = "#override <Btn3Down>: cmdPopupReq()";
static XtActionsRec resizeActions[] = {
                        { "resizeReq", (XtActionProc)resizeCB }
                        };
static XtActionsRec cmdActions[] = {
                        { "cmdPopupReq", (XtActionProc)cmdPopup }
                        };


void createFileSelectionBox( f, type )
  struct FileManager *f;
  int type;
{
  static int firstTime = 1;
  int i, j;
  char dirBuffer[BUFFER_LENGTH], *tokptr, *strptr;
  Widget parent = f->bottomForm;

  /*  parse resize translations  */

  if( firstTime == 1 )
  {
    parsed_resize_xlations = XtParseTranslationTable( resizeTranslations );
    XtAppAddActions( XtWidgetToApplicationContext(parent), resizeActions,
                        XtNumber(resizeActions) );
    parsed_cmd_xlations = XtParseTranslationTable( cmdTranslations );
    XtAppAddActions( XtWidgetToApplicationContext(parent), cmdActions,
                        XtNumber(cmdActions) );
    firstTime = 0;
  }

  /*  create scrolled window and form as child  */

  f->ht = 0;
  f->isManaged = False;

  f->sw = XtVaCreateManagedWidget( "FSBsw",
                xmScrolledWindowWidgetClass, parent,
                XmNshadowThickness, 2,
                XmNtopAttachment, XmATTACH_FORM,
                XmNtopOffset, 5,
                XmNleftAttachment, XmATTACH_FORM,
                XmNleftOffset, 0,
                XmNrightAttachment, XmATTACH_FORM,
                XmNrightOffset, 0,
                XmNbottomAttachment, XmATTACH_FORM,
                XmNbottomOffset, 0,
                XmNvisualPolicy, XmCONSTANT,
                XmNscrollingPolicy, XmAUTOMATIC,
                XmNscrollVertical, False,
                XmNscrollHorizontal, True,
                XmNspacing, 0,
                XmNscrollBarPlacement, XmTOP_RIGHT,
		XmNtranslations, parsed_resize_xlations,
		XmNuserData, (XtPointer)f,
                NULL );

  f->form = XtVaCreateManagedWidget( "FSBform",
		xmRowColumnWidgetClass, f->sw,
                XmNorientation, XmHORIZONTAL,
                XmNpacking, XmPACK_COLUMN,
                XmNresizeHeight, False,
		XmNmarginHeight, 0,
                XmNspacing, 0,
		XmNuserData, (XtPointer)f,
                NULL );

  XtVaSetValues( f->sw, XmNworkWindow, f->form, NULL );
  XtVaGetValues( f->sw, XmNclipWindow, &f->clipWindow,
                 XmNhorizontalScrollBar, &f->sb, NULL );

  f->nDirInfo = 0;
  f->nTotal = 0;

  if( f->type == LOCAL_FILESYSTEM ) 
  {
    struct stat st;

    addListObj( f );
    stat( "/", &st );
    readDir( f, 0, "/", "/", &st, -1, 1, "" );

    memset( (void *)dirBuffer, '\0', sizeof(dirBuffer) );
    strcpy( dirBuffer, getenv("HOME") );
    if( !strcmp( dirBuffer, "" ) )  strcpy( dirBuffer, "/" );

    strptr = dirBuffer;
    i = 0;
    while( (tokptr = strtok(strptr,"/")) != NULL )
    {
      struct DirInfo *di = getIthDir( f->dirInfo, i );
      for( j = 0; di->nFiles; j++ )
      {
        struct FileList *flj = getIthFile( di->fileList, j );
        if( !strcmp( tokptr, flj->name ) )
        {
          XmListSelectPos( di->list, j+1, True );
          XmListSetPos( di->list, j+1 );
          i++;
          break;
        }
      }
      strptr = NULL;
    }
  }
  else
  {
    ftpConnect( f );
  }

  f->isManaged = True;
}

void addListObj( f )
  struct FileManager* f;
{
  struct DirInfo *di;
  Pixel fg, bg;
  Arg args[20];
  int n_args, i;
  Dimension ht;

  di = appendDir( &f->dirInfo );
  di->fileList = (struct FileList *)NULL;
  f->nTotal++;

  XtVaSetValues( f->form,
                 XmNwidth, SL_WIDTH*f->nTotal,
                 NULL );

  if( f->isManaged )
    di->form = XtVaCreateWidget( "form",
                xmFormWidgetClass, f->form,
                XmNresizable, False,
                XmNwidth, SL_WIDTH,
                XmNresizePolicy, XmRESIZE_NONE,
                NULL );
  else
    di->form = XtVaCreateWidget( "form",
                xmFormWidgetClass, f->form,
                XmNresizable, False,
                XmNwidth, SL_WIDTH,
                NULL );

  n_args = 0;
  XtSetArg( args[n_args], XmNtraversalOn, False );  n_args++;
  XtSetArg( args[n_args], XmNshadowThickness, 2 );  n_args++;
  XtSetArg( args[n_args], XmNscrollBarDisplayPolicy, XmSTATIC );  n_args++;
  XtSetArg( args[n_args], XmNselectionPolicy, XmEXTENDED_SELECT );  n_args++;
  XtSetArg( args[n_args], XmNspacing, 0 );  n_args++;
  XtSetArg( args[n_args], XmNtopAttachment, XmATTACH_FORM );  n_args++;
  XtSetArg( args[n_args], XmNtopOffset, 100 );  n_args++;
  XtSetArg( args[n_args], XmNbottomAttachment, XmATTACH_FORM );  n_args++;
  XtSetArg( args[n_args], XmNbottomOffset, 0 );  n_args++;
  XtSetArg( args[n_args], XmNleftAttachment, XmATTACH_FORM );  n_args++;
  XtSetArg( args[n_args], XmNleftOffset, 0 );  n_args++;
  XtSetArg( args[n_args], XmNrightAttachment, XmATTACH_FORM );  n_args++;
  XtSetArg( args[n_args], XmNrightOffset, 0 );  n_args++;
  XtSetArg( args[n_args], XmNlistSizePolicy, XmCONSTANT );  n_args++;
  XtSetArg( args[n_args], XmNscrollBarPlacement, XmBOTTOM_LEFT );  n_args++;
  XtSetArg( args[n_args], XmNuserData, (XtPointer)di );  n_args++;
  di->list = XmCreateScrolledList( di->form, "dirSL", args, n_args );
  XtAddCallback( di->list, XmNextendedSelectionCallback, chooseCB,
        (XtPointer)f );
  XtAddCallback( di->list, XmNmultipleSelectionCallback, chooseCB,
        (XtPointer)f );
  XtAddCallback( di->list, XmNdefaultActionCallback, listItemActivateCB,
	(XtPointer)f );

  XtVaGetValues( di->form, XmNforeground, &fg, XmNbackground, &bg, NULL );

  di->label = XtVaCreateManagedWidget( "",
                xmPushButtonWidgetClass, di->form,
                XmNshadowThickness, 0,
                XmNresizable, False,
                XmNrecomputeSize, False,
                XmNlabelType, XmPIXMAP,
                XmNtopAttachment, XmATTACH_FORM,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, di->list,
                XmNbottomOffset, 25,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNtraversalOn, False,
                XmNforeground, fg,
                XmNbackground, bg,
		XmNuserData, (XtPointer)di,
		XmNtranslations, parsed_cmd_xlations,
                NULL );
  XtAddCallback( di->label, XmNactivateCallback, iconActivateCB, (XtPointer)f );

  di->nameLabel = XtVaCreateManagedWidget( "",
                xmLabelWidgetClass, di->form,
                XmNresizable, False,
                XmNrecomputeSize, False,
                XmNtopAttachment, XmATTACH_WIDGET,
                XmNtopWidget, di->label,
                XmNbottomAttachment, XmATTACH_WIDGET,
                XmNbottomWidget, di->list,
                XmNleftAttachment, XmATTACH_FORM,
                XmNrightAttachment, XmATTACH_FORM,
                XmNforeground, fg,
                XmNbackground, bg,
                NULL );

  XtManageChild( di->list );
  XtManageChild( di->form );

  XFlush( XtDisplay(f->form) );
  XSync( XtDisplay(f->form), False );

  if( f->isManaged )
  {
    XtVaGetValues( f->clipWindow, XmNheight, &ht, NULL );
    if( ht-2 != f->ht )
    {
      f->ht = ht - 2;
      XtVaSetValues( f->form, XmNheight, f->ht, NULL );
    }
  }
}

void resizeCB( w, e )
  Widget w;
  XEvent* e;
{
  XConfigureEvent* event = (XConfigureEvent *)e;
  Dimension ht;
  XtPointer xtp;
  struct FileManager *f;
  int i, max, sliderSize;
 
  if( event->type != ConfigureNotify )  return;

  XtVaGetValues( w, XmNuserData, &xtp, NULL );
  f = (struct FileManager *)xtp;
 
  if( w == f->sw )
  {
    XtVaGetValues( f->clipWindow, XmNheight, &ht, NULL );
    if( f->nTotal > 0 && f->ht != ht-2 )
    {
      f->ht = ht - 2;
      XtVaSetValues( f->form, XmNheight, f->ht, NULL );
    }
  }
}

void chooseCB( w, xtp1, xtp2 )
  Widget w;
  XtPointer xtp1, xtp2;
{
  XmListCallbackStruct *cbs = (XmListCallbackStruct *)xtp2;
  struct FileManager *f = (struct FileManager *)xtp1;

  chooseListItem( w, cbs->selected_item_positions,
        cbs->selected_item_count, f );
}

void sortFiles( d )
  struct DirInfo *d;
{
  int i, j, k;

  for( i = 0; i < d->nFiles-1; i++ )
  {
    struct FileList *fli = getIthFile( d->fileList, i );
    for( j = i+1; j < d->nFiles; j++ )
    {
        int size;
	struct FileList *flj = getIthFile( d->fileList, j );

        if( strlen(fli->name) > strlen(flj->name) )
                size = strlen(flj->name);
        else
                size = strlen(fli->name);

        for( k = 0; k < size; k++ )
        {
          if( fli->name[k] > flj->name[k] )
          {
                char tempString[NAME_LENGTH];
                struct stat s;
                s = fli->s;
                strcpy( tempString, fli->name );
                fli->s = flj->s;
                strcpy( fli->name, flj->name );
                flj->s = s;
                strcpy( flj->name, tempString );
                break;
          }
          else if( fli->name[k] < flj->name[k] )
                break;
        }
    }
  }
}

void forkNexec( w, s, name, runInXterm )
  Widget w;
  struct stat *s;
  char *name;
  int runInXterm;
{
  char cmd[1056];

  memset( (void *)cmd, '\0', sizeof(cmd) );
  getCommand( name, cmd );
  if( !strcmp( cmd, "" ) )  return;

  if( !strcmp( cmd, name ) )
  {
    if( S_ISDIR( s->st_mode ) )
    {
        popupErrorDialog( "Cannot Launch A Directory..." );
        return;
    }
    else if( S_ISLNK( s->st_mode ) )
    {
        popupErrorDialog( "Cannot Launch A Link...\n" );
        return;
    }
    else if( !IS_EXECUTABLE( s->st_mode ) )
    {
        if( getDefaultCommand( name, cmd ) )
	{
	  popupErrorDialog( "No Default Command Is Present..." );
          return;
        }
    }
  }

  if( runInXterm )
  {
    char *ptr = (char *)XtMalloc( strlen(cmd)+1 );
    strcpy( ptr, cmd );
    sprintf( cmd, "xterm -title '%s' -e %s '%s'", ptr, createPausemeFile(), ptr );
    XtFree( ptr );
  }

  launchCmd( w, cmd );
}

void setScrollbar( f, num )
  struct FileManager *f;
  int num;
{
  int max, sliderSize, value;

  XtVaGetValues( f->sb,
                XmNmaximum, &max,
                XmNsliderSize, &sliderSize,
                NULL );

  value = max-sliderSize-(SL_WIDTH*(f->nTotal-f->nDirInfo-num));
  if( value < 0 )  value = 0;
  XmScrollBarSetValues( f->sb, value, sliderSize, SL_WIDTH, SL_WIDTH, True );
}

void accelSelect( di, fmgr )
  DragInfo *di;
  struct FileManager *fmgr;
{
  Widget w = (Widget)NULL;
  int count = 0, *selectedItems;
  char *tokptr, *strptr, *ptr, buffer[256], cmpName[256];

  selectedItems = (int *)XtMalloc( di->numDragNames * sizeof(int) );
  ptr = (char *)XtMalloc( strlen(di->dragName)+1 );
  strcpy( ptr, di->dragName );

  while( sscanf( ptr, "%s", buffer ) == 1 )
  {
    char tmp[256];
    int ctr = 0, j, len = strlen(buffer)+1;

    strcpy( tmp, "/" );
    count++;
    strptr = buffer + 1;
    while( (tokptr = strtok(strptr,"/")) != NULL )
    {
      struct DirInfo *diCtr = getIthDir(fmgr->dirInfo, ctr );

      if( ctr+1 < fmgr->nDirInfo )
        strcpy( cmpName, getIthDir(fmgr->dirInfo,ctr+1)->di.dragName );
      else
        strcpy( cmpName, "" );

      strcat( tmp, tokptr );
      for( j = 0; j < diCtr->nFiles; j++ )
      {
        if( !strcmp( tokptr, getIthFile(diCtr->fileList,j)->name ) )
        {
          if( count == 1 && strcmp( cmpName, tmp ) )
          {
            XmListSelectPos( diCtr->list, j+1, True );
            XmListSetPos( diCtr->list, j+1 );
            XFlush( XtDisplay(fmgr->form) );
            XSync( XtDisplay(fmgr->form), False );
          }
          w = diCtr->list;
          break;
        }
      }
      strptr = NULL;
      ctr++;
      strcat( tmp, "/" );
    }

    selectedItems[count-1] = j+1;
    if( len >= strlen(ptr) )  break;
    ptr += len;
  }

  XmListDeselectAllItems( w );
  XtVaSetValues( w, XmNselectionPolicy, XmMULTIPLE_SELECT, NULL );
  for( count = 0; count < di->numDragNames; count++ )
    XmListSelectPos( w, selectedItems[count], False );
  XtVaSetValues( w, XmNselectionPolicy, XmEXTENDED_SELECT, NULL );
  chooseListItem( w, selectedItems, di->numDragNames, fmgr );
  if( di->numDragNames > 0 )  XmListSetPos( w, selectedItems[0] );

  XtFree( (char *)ptr );
  XtFree( (char *)selectedItems );
}

void deleteDirStruct( from, f )
  int from;
  struct FileManager *f;
{
  int i;
  for( i = from; i < f->nDirInfo; i++ )
  {
    struct DirInfo *dI = getIthDir( f->dirInfo, i );
    XmString xms = XmStringCreateSimple( "" );

    if( dI->nFiles > 0 ) removeAllFiles( &dI->fileList );
    dI->nFiles = 0;
    if( dI->di.dragName ) XtFree( dI->di.dragName );
    XtVaSetValues( dI->label,
                XmNlabelPixmap, getPixmap( "BLANK", dI->s, f->type, "" ),
                NULL );

    XmDropSiteUnregister( dI->label );
    unregisterDragSite( dI->label );

    XtVaSetValues( dI->nameLabel, XmNlabelString, xms, NULL );
    XmListDeleteAllItems( dI->list );
    XmStringFree( xms );
  }
  f->nDirInfo = from;
  setScrollbar( f, 0 );
}

void listItemActivateCB( w, xtp1, xtp2 )
  Widget w;
  XtPointer xtp1, xtp2;
{
  XtPointer xtp;
  struct stat s;
  int rix = 0;
  char *ptr;
  XmListCallbackStruct *cbs = (XmListCallbackStruct *)xtp2;
  struct FileManager *f = (struct FileManager *)xtp1;
  struct DirInfo *dInfo;
  struct FileList *fl;

  setWatchCursor();
  XtVaGetValues( w, XmNuserData, &xtp, NULL );
  dInfo = (struct DirInfo *)xtp;
  fl = getIthFile( dInfo->fileList, cbs->item_position-1 );
  ptr = (char *)XtMalloc( strlen(dInfo->di.dragName) +
	strlen(fl->name) + 10 );
  sprintf( ptr, "%s/%s", dInfo->di.dragName, fl->name );

  if( f->runInXterm )  rix = 1;
  forkNexec( w, &fl->s, ptr, rix );
  unsetWatchCursor();
  XtFree( ptr );
}

void iconActivateCB( w, xtp1, xtp2 )
  Widget w;
  XtPointer xtp1, xtp2;
{
  XmPushButtonCallbackStruct *cbs = (XmPushButtonCallbackStruct *)xtp2;
  struct FileManager *f = (struct FileManager *)xtp1;
  struct DirInfo *dInfo = f->dirInfo;
  int i;

  for( i = 0; i < f->nDirInfo; i++ )  
  {
    if( w == dInfo->label )  
	break;
    else
 	dInfo = dInfo->next;
  }
  if( i >= f->nDirInfo )  return;

  if( cbs->click_count == 2 && f->type == LOCAL_FILESYSTEM )
  {
    XtPointer xtp;
    struct stat s;
    int rix = 0;

    setWatchCursor();
    XtVaGetValues( w, XmNuserData, &xtp, NULL );
    dInfo = (struct DirInfo *)xtp;
    if( stat( dInfo->di.dragName, &s ) < 0 )
    {
      char msg[256];
      sprintf( msg, "stat() failed for %s", dInfo->di.dragName );
      popupErrorDialog( msg );
      unsetWatchCursor();
      return;
    }
    if( f->runInXterm )  rix = 1;
    forkNexec( w, &s, dInfo->di.dragName, rix );
    unsetWatchCursor();
  }
  else if( cbs->click_count == 1 )
  {
    setWatchCursor();
    XmListDeselectAllItems( dInfo->list );
    XmListSetPos( dInfo->list, 1 );
    deleteDirStruct( i+1, f );
    unsetWatchCursor();
  }
}

void cmdPopup( w, e )
  Widget w;
  XEvent* e;
{
  XtPointer xtp;
  struct DirInfo *dInfo;
  char cmd[1056];

  XtVaGetValues( w, XmNuserData, &xtp, NULL );
  dInfo = (struct DirInfo *)xtp;
  memset( (void *)cmd, '\0', sizeof(cmd) );
  getCommand( dInfo->di.dragName, cmd );

  if( !strcmp( dInfo->di.dragName, cmd ) )
  {
    if(  !IS_EXECUTABLE( dInfo->s.st_mode ) )
    {
      if( getDefaultCommand( dInfo->di.dragName, cmd ) )
      {
        popupErrorDialog( "No Default Command Is Present..." );
        return;
      }
    }
  }

  if( strcmp( cmd, "" ) )  popupCmdWithCmd( w, cmd );
}

void reloadDir( f, index )
  struct FileManager *f;
  int index;
{
  char dirBuffer[BUFFER_LENGTH];
  DIR *dp = (DIR *)NULL;
  struct dirent *dirp;
  int j;
  XmString *xms;
  struct DirInfo *fd = getIthDir( f->dirInfo, index );
  struct FileList *fl;

  if( f->type == LOCAL_FILESYSTEM )
  {
    if( stat( fd->di.dragName, &fd->s ) < 0 )
    {
       fprintf( stderr, "Cannot stat %s\n", fd->di.dragName );
       return;
    }

    if( (dp = opendir( fd->di.dragName )) == (DIR *)NULL )  return;
  }

  if( fd->nFiles > 0 )
  {
    fd->nFiles = 0;
    removeAllFiles( &fd->fileList );
  }

  if( f->type == LOCAL_FILESYSTEM )
  {
    while( (dirp = readdir(dp)) != NULL )
    {
      int ct = fd->nFiles;
      struct stat s;

      if( !strcmp( dirp->d_name, "." ) || !strcmp( dirp->d_name, ".." ) )  continue;

      if( !strcmp( fd->di.dragName, "/" ) )
        sprintf( dirBuffer, "/%s", dirp->d_name );
      else
        sprintf( dirBuffer, "%s/%s", fd->di.dragName, dirp->d_name );

      /* fl->s = NULL; */
      if( stat( dirBuffer, &s ) < 0 ) continue;

      fl = appendFile( &fd->fileList );
      fl->s = s;
      strcpy( fl->name, dirp->d_name );
      fd->nFiles++;
    }

    closedir( dp );
    sortFiles( fd );

    fl = fd->fileList;
    xms = (XmString *)XtMalloc( fd->nFiles * sizeof(XmString) );
    for( j = 0; j < fd->nFiles; j++ )
    {
        strcpy( dirBuffer, fl->name );
        if( S_ISDIR( fl->s.st_mode ) )
           strcat( dirBuffer, "/" );
        else if( S_ISLNK( fl->s.st_mode ) )
           strcat( dirBuffer, "@" );
        else if( IS_EXECUTABLE( fl->s.st_mode ) )
           strcat( dirBuffer, "*" );
        xms[j] = XmStringCreateSimple( dirBuffer );
	fl = fl->next;
    }

    XmListDeleteAllItems( fd->list );
    XtVaSetValues( fd->list,
                XmNitems, xms,
                XmNitemCount, fd->nFiles,
                NULL );

    for( j = 0; j < fd->nFiles; j++ )
        XmStringFree( xms[j] );
    XtFree( (char *)xms );
  }
  else
  {
    sprintf( dirBuffer, "LIST %s", fd->di.dragName );
    ftpSendCommand( f, dirBuffer, index, "" );
  }
}

void readDir( f, index, s, l, st, pIndex, num, perm )
  struct FileManager *f;
  int index, pIndex, num;
  char *s, *l;
  struct stat* st;
  char *perm;
{
  XmString labelString;
  struct DirInfo *fd = getIthDir( f->dirInfo, index );

  fd->di.numDragNames = num;
  fd->nFiles = 0;
  strcpy( fd->di.dirName, l );
  fd->di.dragName = (char *)XtMalloc( strlen(s)+1 );
  strcpy( fd->di.dragName, s );

  if( num > 1 )
    fd->di.p = getPixmap( "MULTIPLE", *st, f->type, "" );
  else
    fd->di.p = getPixmap( s, *st, f->type, perm );

  if( fd->di.p )
    XtVaSetValues( fd->label, XmNlabelPixmap, fd->di.p, NULL );

  labelString = XmStringCreateSimple( l );
  XtVaSetValues( fd->nameLabel, XmNlabelString, labelString, NULL );
  XmStringFree( labelString );

  registerDragSite( fd->label );
  unregisterDragSite( fd->nameLabel );
  if( num == 1 && f->type == LOCAL_FILESYSTEM && S_ISDIR( st->st_mode ) )
      registerDropSite( fd->label, DIR_DROP_SITE );

  f->nDirInfo++;
  if( !strcmp( s, "" ) )  return;
  if( num == 1 )  reloadDir( f, index );
}

void chooseListItem( w, sip, sic, f )
  Widget w;
  int *sip, sic;
  struct FileManager *f;
{
  int i;
  struct DirInfo *di = f->dirInfo;

  for( i = 0; i < f->nDirInfo; i++ )
  {
    if( w == di->list )
    {
        setWatchCursor();
        loadDirStruct( f, i, sip, sic );
        unsetWatchCursor();
        break;
    }
    di = di->next;
  }
}

void loadDirStruct( f, whichList, chosenItems, nChosenItems )
  struct FileManager *f;
  int *chosenItems, nChosenItems, whichList;
{
  char dirBuffer[BUFFER_LENGTH], labelName[NAME_LENGTH];
  int i, j;
  struct stat sTmp;
  struct DirInfo *di = getIthDir( f->dirInfo, whichList+1 );

  for( i = whichList+1; i < f->nDirInfo; i++ )
  {
    struct stat st;
    XmString xms = XmStringCreateSimple( "" );

    if( di->nFiles > 0 ) removeAllFiles( &di->fileList );
    di->nFiles = 0;
    if( (char *)di->di.dragName ) XtFree( (char *)di->di.dragName );
    XtVaSetValues( di->label,
                XmNlabelPixmap, getPixmap( "BLANK", st, f->type, "" ),
                NULL );

    XmDropSiteUnregister( di->label );
    unregisterDragSite( di->label );

    XtVaSetValues( di->nameLabel,
                XmNlabelString, xms,
                NULL );
    XmListDeleteAllItems( di->list );
    XmStringFree( xms );
    di = di->next;
  }

  f->nDirInfo = whichList + 1;
  di = getIthDir( f->dirInfo, whichList );

  if( nChosenItems > 1 )
  {
    char tmp[256];
    sprintf( labelName, "%d Items Selected", nChosenItems );
    strcpy( dirBuffer, "" );
    strcpy( tmp, "" );
    if( whichList > 0 )  strcat( tmp, di->di.dragName );
    strcat( tmp, "/" );

    for( i = 0; i < nChosenItems; i++ )
    {
      struct FileList *fl = getIthFile( di->fileList, chosenItems[i]-1 );
      if( i > 0 )  strcat( dirBuffer, " " );
      strcat( dirBuffer, tmp );
      strcat( dirBuffer, fl->name );
    }
  }
  else
  {
    struct FileList *fl = getIthFile( di->fileList, chosenItems[0]-1 );
    strcpy( labelName, fl->name );
    strcpy( dirBuffer, "" );
    if( whichList > 0 )  strcat( dirBuffer, di->di.dragName );
    strcat( dirBuffer, "/" );
    strcat( dirBuffer, fl->name );
  }

  if( f->nDirInfo >= f->nTotal )  addListObj( f );
  setScrollbar( f, 1 );

  if( nChosenItems == 1 )
  {
     struct FileList *fl = getIthFile( di->fileList, chosenItems[0]-1 );
     sTmp = fl->s;
     readDir( f, f->nDirInfo, dirBuffer, labelName, &sTmp, 
	chosenItems[0]-1, 1, fl->perm );
  }
  else
  {
     readDir( f, f->nDirInfo, dirBuffer, labelName, &sTmp, -1, 
	nChosenItems, "" );
  }
}

void updateFileManager( f )
  struct FileManager *f;
{
  int j, isSel = 1;

  if( f->type != LOCAL_FILESYSTEM )  return;

  for( j = 0; j < f->nDirInfo; j++ )
  {
    struct DirInfo *fd = getIthDir( f->dirInfo, j );
    char dirBuffer[BUFFER_LENGTH];
    DIR *dp = (DIR *)NULL;
    struct dirent *dirp;
    struct stat s;
    XmString *chosenItem, *xms;
    int nChosenItems = 0, iii, set = 1;
    struct FileList *fl;
    int *pos_list, pos_count = 0;

    if( isSel == 0 )
    {
      deleteDirStruct( j, f );
      break;
    }
    else if( fd->di.numDragNames > 1 )
    {
      break;
    }

    if( stat( fd->di.dragName, &s ) < 0 )
    {
      /* directory does not exist anymore so delete all subsequent directories */ 
      deleteDirStruct( j, f );
      break;
    }

    /* check to see if directory has changed; if not, continue */
    if( fd->s.st_mtime == s.st_mtime && fd->s.st_ctime == s.st_ctime )  continue;

    /* store selected items */
    if( XmListGetSelectedPos( fd->list, &pos_list, &pos_count ) )
    {
      chosenItem = (XmString *)XtMalloc( pos_count * sizeof(XmString) );
      for( iii = 0; iii < pos_count; iii++ )
        chosenItem[iii] = XmStringCreateSimple( 
		getIthFile(fd->fileList,pos_list[iii]-1)->name );
      nChosenItems = pos_count;
    }

    if( (dp = opendir( fd->di.dragName )) == (DIR *)NULL )  
    {
      /* permissions may have changed so that we can no longer read this
         directory so delete all subsequent directories */
      deleteDirStruct( j, f );
      break;
    }

    if( fd->nFiles > 0 )
    {
      fd->nFiles = 0;
      removeAllFiles( &fd->fileList );
    }

    while( (dirp = readdir(dp)) != NULL )
    {
      struct stat tmpStat;
      int ct = fd->nFiles;

      if( !strcmp( dirp->d_name, "." ) || !strcmp( dirp->d_name, ".." ) )  continue;

      if( !strcmp( fd->di.dragName, "/" ) )
        sprintf( dirBuffer, "/%s", dirp->d_name );
      else
        sprintf( dirBuffer, "%s/%s", fd->di.dragName, dirp->d_name );

      /* fl->s = NULL; */
      if( stat( dirBuffer, &tmpStat ) < 0 )  continue;

      fl = appendFile( &fd->fileList );
      fl->s = tmpStat;
      strcpy( fl->name, dirp->d_name );
      fd->nFiles++;
    }

    closedir( dp );
    sortFiles( fd );

    fl = fd->fileList;
    xms = (XmString *)XtMalloc( fd->nFiles * sizeof(XmString) );
    for( iii = 0; iii < fd->nFiles; iii++ )
    {
        strcpy( dirBuffer, fl->name );
        if( S_ISDIR( fl->s.st_mode ) )
           strcat( dirBuffer, "/" );
        else if( S_ISLNK( fl->s.st_mode ) )
           strcat( dirBuffer, "@" );
        else if( IS_EXECUTABLE( fl->s.st_mode ) )
           strcat( dirBuffer, "*" );
        xms[iii] = XmStringCreateSimple( dirBuffer );
        fl = fl->next;
    }

    XmListDeleteAllItems( fd->list );
    XtVaSetValues( fd->list,
                XmNitems, xms,
                XmNitemCount, fd->nFiles,
                NULL );

    isSel = 1;
    XtVaSetValues( fd->list, XmNselectionPolicy, XmMULTIPLE_SELECT, NULL );
    for( iii = 0; iii < nChosenItems; iii++ )
    {
      int pos;

      fl = fd->fileList;
      for( pos = 0; pos < fd->nFiles; pos++ )
      {
        XmString tmpXms = XmStringCreateSimple( fl->name );
	if( XmStringCompare( tmpXms, chosenItem[iii] ) ) 
	{
	  XmStringFree( tmpXms );
	  break;
	}
	fl = fl->next;
        XmStringFree( tmpXms );
      }

      if( pos < fd->nFiles )
      {
        XmListSelectPos( fd->list, pos, False );
        if( set == 1 ) 
        {
          XmListSetPos( fd->list, pos );
          set = 0;
        }
      }
      else
	isSel = 0;
    }
    XtVaSetValues( fd->list, XmNselectionPolicy, XmEXTENDED_SELECT, NULL );

    if( fd->nFiles > 0 )
    {
      for( iii = 0; iii < fd->nFiles; iii++ )
        XmStringFree( xms[iii] );
      XtFree( (char *)xms );
    }

    if( nChosenItems > 0 )
    {
      for( iii = 0; iii < nChosenItems; iii++ )
        XmStringFree( chosenItem[iii] );
      XtFree( (char *)chosenItem );
    }
  }
}
